/*
  vt1xx terminal emulation by
  Rene Rebe <rene@exactcode.de>
  
  Lossy based on code of bogl-term by:
  Edmund GRIMLEY EVANS <edmundo@rano.org>
  
  Additional information from the Linux kernel driver/char/vt.c by:
  Linus Torvalds
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA. */

#define _XOPEN_SOURCE 500
#include <errno.h>
#include <fcntl.h>
#include <locale.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <unistd.h>

#include "eTerm.hh"

#include "EvasHelper.hh"
#include "Evas_Engine_FB.h"

#include <linux/fb.h>

#include <iostream>

static int child_pid = 0;
static struct termios ttysave;

#include "TermUtil.hh"

static int quit = 0;

void sigterm (int sig)
{
  quit = 1;
}

// screen width / height
int sw = 80;
int sh = 25;

int main (int argc, char *argv[])
{
  struct termios ntio;
  int ret;
  char buf[8192];
  struct timeval tv;
  int ptyfd, ttyfd;
  char *locale = "", *command = NULL;
  int i;
  char o = ' ';
  int pending = 0;
  
  int win_w = 1;
  int win_h = 1;
  
  // evas
  EvasCanvas* evas;
  eTerm* term;

  
  struct fb_var_screeninfo fb_var;
  int fb;
  
  fb = open ("/dev/fb0", O_RDWR);
  if (fb < 0) {
    perror ("fb");
    return 1;
  }
  
  if (ioctl(fb, FBIOGET_VSCREENINFO, &fb_var) == -1) {
    perror ("fb");
    return 1;
  }
  close (fb);
  win_w = fb_var.xres;
  win_h = fb_var.yres;
  
  evas = new EvasCanvas ();
  evas->OutputMethod ("fb");
  
  evas->OutputSize (win_w, win_h);
  evas->OutputViewport (0, 0, win_w, win_h);
  
  Evas_Engine_Info_FB* einfo;
  einfo = (Evas_Engine_Info_FB*) evas->EngineInfo ();
  
  /* the following is specific to the engine */
  einfo->info.virtual_terminal = 0;
  einfo->info.device_number = 0;
  einfo->info.refresh = 0;
  einfo->info.rotation = 0;
  
  evas->EngineInfo ( (Evas_Engine_Info*)einfo );
  
  /* Setup */
  evas->FontPathPrepend ("/usr/X11/lib/X11/fonts/TTF/");
  evas->FontPathPrepend ("/usr/X11/lib/X11/fonts/TrueType/");
  evas->FontPathPrepend ("/usr/X11/share/fonts/TTF/");
  evas->FontPathPrepend ("/opt/e17/share/evas/data/");
  
  if (true) {
    evas->ImageCache (1024 * 1024);
    evas->FontCache (256 * 1024);
  }
  else {
    evas->ImageCache (0);
    evas->FontCache (0);
  }
  
  term = new eTerm (evas, sw, sh);
  sw = win_w / term->FontWidth();
  sh = win_h / term->FontHeight();
  term->Resize (sw, sh);
  
  for (i = 1 ; i < argc ; ++i)
    if (argv[i][0] == '-')
      switch (argv[i][1])
	{
	case 'f':
	case 'l':
	  o = argv[i][1];
	  break;

	default:
	  printf ("unknown option: %c\n", argv[i][1]);
	}
    else
      switch (o)
	{
	case ' ':
	  command = argv[i];
	  break;

	case 'l':
	  locale = argv[i];
	  o = ' ';
	  break;
	}

  setlocale (LC_CTYPE, locale);
  
  tcgetattr (0, &ttysave);

  if (get_ptytty (&ptyfd, &ttyfd)) {
    perror("can't get a pty");
    exit(1);
  }
  
  spawn_shell (ptyfd, ttyfd, command == NULL ? "/bin/sh" : command);
  
  evas->Render ();
  
  // signal(SIGHUP, reload_font);
  signal (SIGTERM, sigterm);

  ntio = ttysave;
  ntio.c_lflag &= ~(ECHO|ISIG|ICANON|XCASE);
  ntio.c_iflag = 0;
  ntio.c_oflag &= ~OPOST;
  ntio.c_cc[VMIN] = 1;
  ntio.c_cc[VTIME] = 0;
  ntio.c_cflag |= CS8;
  ntio.c_line = 0;
  tcsetattr (0, TCSAFLUSH, &ntio);

  set_window_size (ttyfd, sw, sh);

  for (;;) {
    fd_set fds;
    int max = 0;

    if (quit)
      break;
    
    if (pending)
      {
    	tv.tv_sec = 0;
    	tv.tv_usec = 0;
      }
    else
      {
    	tv.tv_sec = 10;
    	tv.tv_usec = 100000;
      }
    FD_ZERO(&fds);
    FD_SET(0, &fds);
    FD_SET(ptyfd, &fds);
    if (ptyfd > max)
      max = ptyfd;
    ret = select (max+1, &fds, NULL, NULL, &tv);
    
    if (quit)
      break;
    
    if (ret == 0 || (ret < 0 && errno == EINTR))
      {
	if (pending)
	  {
	    pending = 0;
	    
	    term->redraw ();
	    evas->Render ();
	    usleep (10000);
	  }
      }
    
    if (ret < 0)
      perror("select");
    if (FD_ISSET(0, &fds)) {
      ret = read(0, buf, sizeof(buf));
      if (ret > 0)
	write (ptyfd, buf, ret);
    }
    else if (FD_ISSET(ptyfd,&fds)) {
      ret = read(ptyfd, buf, sizeof(buf));
      if (ret > 0)
	{
	  // output to screen
	  term->out (buf, ret);
	  pending = 1;
	}
    }
  }

  delete evas;
  evas = 0;  
  
  return 0;
}
